import typing

from pydantic import BaseModel

from mlfoundry.log_types.whylogs_types.summary import DataFrameSummary


class TabularDatasetStats(BaseModel):
    features: DataFrameSummary
    actuals: typing.Optional[DataFrameSummary] = None
    predictions: typing.Optional[DataFrameSummary] = None

    @property
    def log_type(self) -> str:
        return "dataset-summary-v1"



import logging

import pandas as pd
from sklearn.metrics import (
    explained_variance_score,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
)

from mlfoundry.constants import (
    MULTI_DIMENSIONAL_METRICS,
    NON_MULTI_DIMENSIONAL_METRICS,
    PROB_MULTI_DIMENSIONAL_METRICS,
    PROB_NON_MULTI_DIMENSIONAL_METRICS,
)
from mlfoundry.metrics.v1.base_metrics import BaseMetrics

log = logging.getLogger(__name__)


class TimeSeriesMetrics(BaseMetrics):
    TYPE_OF_MODEL = "timeseries"
    MEAN_SQUARED_ERROR = "mean_squared_error"
    MEAN_ABSOLUTE_ERROR = "mean_absolute_error"
    R2_SCORE = "r2_score"
    ROOT_MEAN_SQUARE_ERROR = "root_mean_squared_error"

    def __init__(self):
        super().__init__()
        self.timeseries_metrics = {
            NON_MULTI_DIMENSIONAL_METRICS: {
                self.MEAN_SQUARED_ERROR: self.compute_mean_squared_error,
                self.MEAN_ABSOLUTE_ERROR: self.compute_mean_absolute_error,
                self.R2_SCORE: self.compute_r2_score,
                self.ROOT_MEAN_SQUARE_ERROR: self.compute_root_mean_squared_error,
            },
            MULTI_DIMENSIONAL_METRICS: {},
            PROB_NON_MULTI_DIMENSIONAL_METRICS: {},
            PROB_MULTI_DIMENSIONAL_METRICS: {},
        }

    def compute_metrics(
        self, features_df, predictions, labels, pred_probabilities=None
    ):
        """
        Compute metrics for timeseries models.This method calls the parent class as well for any supported metrics.
        :param features_df:
        :param predictions:
        :param labels:
        :return:
            metrics_dict
        """
        metrics_from_parent = super().compute_metrics(
            features_df, predictions, labels, pred_probabilities
        )

        all_metrics = metrics_from_parent
        for metric in self.timeseries_metrics[NON_MULTI_DIMENSIONAL_METRICS].keys():
            all_metrics[NON_MULTI_DIMENSIONAL_METRICS][
                metric
            ] = self.timeseries_metrics[NON_MULTI_DIMENSIONAL_METRICS][metric](
                labels, predictions
            )
        for metric in self.timeseries_metrics[MULTI_DIMENSIONAL_METRICS].keys():
            all_metrics[MULTI_DIMENSIONAL_METRICS][metric] = self.timeseries_metrics[
                MULTI_DIMENSIONAL_METRICS
            ][metric](labels, predictions)

        if pred_probabilities:
            for metric in self.timeseries_metrics[
                PROB_NON_MULTI_DIMENSIONAL_METRICS
            ].keys():
                all_metrics[NON_MULTI_DIMENSIONAL_METRICS][
                    metric
                ] = self.timeseries_metrics[PROB_NON_MULTI_DIMENSIONAL_METRICS][metric](
                    labels, pred_probabilities
                )
            for metric in self.timeseries_metrics[
                PROB_MULTI_DIMENSIONAL_METRICS
            ].keys():
                all_metrics[MULTI_DIMENSIONAL_METRICS][
                    metric
                ] = self.timeseries_metrics[PROB_MULTI_DIMENSIONAL_METRICS][metric](
                    labels, pred_probabilities
                )

        return all_metrics

    @staticmethod
    def compute_mean_squared_error(y_true, y_pred):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_pred:
        :return:
        """
        return mean_squared_error(y_true, y_pred)

    @staticmethod
    def compute_mean_absolute_error(y_true, y_pred):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_pred:
        :return:
        """
        return mean_absolute_error(y_true, y_pred)

    @staticmethod
    def compute_r2_score(y_true, y_pred):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_pred:
        :return:
        """
        return r2_score(y_true, y_pred)

    @staticmethod
    def compute_root_mean_squared_error(y_true, y_pred):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_pred:
        :return:
        """
        return mean_squared_error(y_true, y_pred, squared=False)



import fnmatch
import os
import posixpath
import tempfile
import typing

import pandas as pd
from mlflow.tracking import MlflowClient
from pydantic import BaseModel
from whylogs import DatasetProfile

from mlfoundry.dataset.types import Profiles, Schema
from mlfoundry.enums import FileFormat
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types import TabularDatasetStats

DATASET_PREFIX = os.path.join("mlf", "tabular-dataset")


def _save_dataframe(
    local_dir: str,
    entity_name: str,
    dataframe: pd.DataFrame,
    file_format: FileFormat,
) -> str:
    file_name = f"{entity_name}.{file_format.value}"
    file_path = os.path.join(local_dir, file_name)

    if file_format is FileFormat.PARQUET:
        dataframe.to_parquet(file_path, index=False)
    elif file_format is FileFormat.CSV:
        dataframe.to_csv(file_path, index=False)
    else:
        raise MlFoundryException(f"cannot save dataframe in {file_format} file format")

    return file_path


def _load_dataframe(local_file_path: str) -> pd.DataFrame:
    _, file_ext = os.path.splitext(local_file_path)
    if len(file_ext) <= 1:
        raise MlFoundryException(
            f"cannot recognize file format, '{file_ext}' is not valid"
        )
    # ".csv" -> "csv"
    file_ext = file_ext[1:]
    file_format = FileFormat(file_ext)

    if file_format is FileFormat.CSV:
        return pd.read_csv(local_file_path)
    elif file_format is FileFormat.PARQUET:
        return pd.read_parquet(local_file_path)
    else:
        raise MlFoundryException(
            f"cannot load dataframe from {file_format} file format"
        )


def _save_series(
    local_dir: str,
    entity_name: str,
    series: pd.Series,
    file_format: FileFormat,
):
    dataframe = series.to_frame()
    return _save_dataframe(
        local_dir=local_dir,
        entity_name=entity_name,
        dataframe=dataframe,
        file_format=file_format,
    )


def _load_series(local_file_path: str) -> pd.Series:
    dataframe = _load_dataframe(local_file_path)
    return dataframe.iloc[:, 0]


def _save_pydantic_as_json(local_dir: str, file_name: str, model: BaseModel):
    with open(os.path.join(local_dir, file_name), "w") as fp:
        fp.write(model.json())


class TabularDatasetSerDe:
    _FEATURE_ENTITY_NAME = "features"
    _PREDICTIONS_ENTITY_NAME = "predictions"
    _ACTUALS_ENTITY_NAME = "actuals"
    _SCHEMA_FILE_NAME = "schema.json"
    _STATS_FILE_NAME = "stats.json"
    _FEATURES_PROFILE_FILE_NAME = "features_profile.pb.bin"
    _PREDICTIONS_PROFILE_FILE_NAME = "predictions_profile.pb.bin"
    _ACTUALS_PROFILE_FILE_NAME = "actuals_profile.pb.bin"

    def __init__(self, mlflow_client: MlflowClient):
        self.mlflow_client: MlflowClient = mlflow_client
        self.dataset_prefix = DATASET_PREFIX

    def _get_artifact_dir(self, dataset_name: str) -> str:
        return os.path.join(self.dataset_prefix, dataset_name)

    def _get_versioned_artifact_dir(self, dataset_name: str, version: str) -> str:
        return os.path.join(self._get_artifact_dir(dataset_name), version)

    def get_latest_version(self, run_id: str, dataset_name: str) -> str:
        dataset_artifact_dir = self._get_artifact_dir(dataset_name)
        number_of_versions_present = len(
            self.mlflow_client.list_artifacts(run_id=run_id, path=dataset_artifact_dir)
        )
        return str(number_of_versions_present)

    def _get_artifact_dir_new_version(
        self, run_id: str, dataset_name: str
    ) -> typing.Tuple[str, str]:
        last_version = self.get_latest_version(run_id=run_id, dataset_name=dataset_name)
        version = str(int(last_version) + 1)
        return (
            self._get_versioned_artifact_dir(dataset_name, version=version),
            version,
        )

    def _get_latest_dataset_artifact_dir(self, run_id: str, dataset_name: str) -> str:
        latest_version = self.get_latest_version(
            run_id=run_id, dataset_name=dataset_name
        )
        return self._get_versioned_artifact_dir(dataset_name, version=latest_version)

    def _list_artifacts_with_pattern(
        self, run_id: str, path: str, pattern: str
    ) -> typing.List:
        return [
            file_info
            for file_info in self.mlflow_client.list_artifacts(run_id=run_id, path=path)
            if fnmatch.fnmatch(file_info.path, pattern)
        ]

    def _save_profiles(self, local_dir, profiles: Profiles):
        features_profile_path = os.path.join(
            local_dir, TabularDatasetSerDe._FEATURES_PROFILE_FILE_NAME
        )
        profiles.features.write_protobuf(features_profile_path)

        if profiles.predictions is not None:
            predictions_profile_path = os.path.join(
                local_dir, TabularDatasetSerDe._PREDICTIONS_PROFILE_FILE_NAME
            )
            profiles.predictions.write_protobuf(predictions_profile_path)

        if profiles.actuals is not None:
            actuals_profile_path = os.path.join(
                local_dir, TabularDatasetSerDe._ACTUALS_PROFILE_FILE_NAME
            )
            profiles.actuals.write_protobuf(actuals_profile_path)

    def _load_file(
        self,
        run_id: str,
        dataset_name: str,
        version: str,
        file_name: str,
        deserializer: typing.Callable,
    ):
        dataset_artifact_dir = self._get_versioned_artifact_dir(
            dataset_name, version=version
        )
        file_path = os.path.join(dataset_artifact_dir, file_name)
        artifacts = self._list_artifacts_with_pattern(
            run_id=run_id, path=dataset_artifact_dir, pattern=file_path
        )
        if len(artifacts) == 0:
            return None
        artifact_path = artifacts[0].path
        with tempfile.TemporaryDirectory() as local_dir:
            local_path = self.mlflow_client.download_artifacts(
                run_id, artifact_path, local_dir
            )
            return deserializer(local_path)

    def _load_entity(
        self,
        run_id: str,
        dataset_name: str,
        entity_name: str,
        version: str,
        deserializer: typing.Callable,
    ):
        dataset_artifact_dir = self._get_versioned_artifact_dir(
            dataset_name, version=version
        )
        # for features, predictions, actuals we may either save it as parquet
        # or CSV. Either feature.csv or feature.parquet.
        # So here we just search a pattern like `feature.*` to
        # get the artifact, then deserializer will take care of the rest
        pattern = posixpath.join(dataset_artifact_dir, f"{entity_name}.*")
        artifacts = self._list_artifacts_with_pattern(
            run_id=run_id, path=dataset_artifact_dir, pattern=pattern
        )
        if len(artifacts) == 0:
            return None

        artifact_path = artifacts[0].path
        with tempfile.TemporaryDirectory() as local_dir:
            local_path = self.mlflow_client.download_artifacts(
                run_id, artifact_path, local_dir
            )
            return deserializer(local_path)

    def save_dataset(
        self,
        run_id: str,
        dataset_name: str,
        features: pd.DataFrame,
        predictions: typing.Optional[pd.Series],
        actuals: typing.Optional[pd.Series],
        schema: Schema,
        stats: TabularDatasetStats,
        profiles: Profiles,
        file_format: FileFormat,
        only_stats: bool,
    ) -> str:
        with tempfile.TemporaryDirectory() as local_dir:
            if not only_stats:
                _save_dataframe(
                    local_dir=local_dir,
                    entity_name=TabularDatasetSerDe._FEATURE_ENTITY_NAME,
                    dataframe=features,
                    file_format=file_format,
                )

                if predictions is not None:
                    _save_series(
                        local_dir=local_dir,
                        entity_name=TabularDatasetSerDe._PREDICTIONS_ENTITY_NAME,
                        series=predictions,
                        file_format=file_format,
                    )
                if actuals is not None:
                    _save_series(
                        local_dir=local_dir,
                        entity_name=TabularDatasetSerDe._ACTUALS_ENTITY_NAME,
                        series=actuals,
                        file_format=file_format,
                    )

            _save_pydantic_as_json(
                local_dir=local_dir,
                file_name=TabularDatasetSerDe._SCHEMA_FILE_NAME,
                model=schema,
            )
            _save_pydantic_as_json(
                local_dir=local_dir,
                file_name=TabularDatasetSerDe._STATS_FILE_NAME,
                model=stats,
            )
            self._save_profiles(local_dir=local_dir, profiles=profiles)

            artifact_path, version = self._get_artifact_dir_new_version(
                run_id=run_id, dataset_name=dataset_name
            )
            self.mlflow_client.log_artifacts(
                run_id=run_id, local_dir=local_dir, artifact_path=artifact_path
            )
            return version

    def load_schema(self, run_id: str, dataset_name: str, version: str) -> Schema:
        return self._load_file(
            run_id=run_id,
            dataset_name=dataset_name,
            version=version,
            file_name=TabularDatasetSerDe._SCHEMA_FILE_NAME,
            deserializer=Schema.parse_file,
        )

    def load_stats(
        self, run_id: str, dataset_name: str, version: str
    ) -> TabularDatasetStats:
        return self._load_file(
            run_id=run_id,
            dataset_name=dataset_name,
            version=version,
            file_name=TabularDatasetSerDe._STATS_FILE_NAME,
            deserializer=TabularDatasetStats.parse_file,
        )

    def load_profiles(self, run_id: str, dataset_name: str, version: str) -> Profiles:
        features_profile = self._load_file(
            run_id=run_id,
            dataset_name=dataset_name,
            version=version,
            file_name=TabularDatasetSerDe._FEATURES_PROFILE_FILE_NAME,
            deserializer=DatasetProfile.read_protobuf,
        )
        actuals_profile = self._load_file(
            run_id=run_id,
            dataset_name=dataset_name,
            version=version,
            file_name=TabularDatasetSerDe._ACTUALS_PROFILE_FILE_NAME,
            deserializer=DatasetProfile.read_protobuf,
        )
        predictions_profile = self._load_file(
            run_id=run_id,
            dataset_name=dataset_name,
            version=version,
            file_name=TabularDatasetSerDe._PREDICTIONS_PROFILE_FILE_NAME,
            deserializer=DatasetProfile.read_protobuf,
        )
        return Profiles(
            features=features_profile,
            actuals=actuals_profile,
            predictions=predictions_profile,
        )

    def load_features(
        self, run_id: str, dataset_name: str, version: str
    ) -> typing.Optional[pd.DataFrame]:
        return self._load_entity(
            run_id=run_id,
            dataset_name=dataset_name,
            entity_name=TabularDatasetSerDe._FEATURE_ENTITY_NAME,
            deserializer=_load_dataframe,
            version=version,
        )

    def load_actuals(
        self, run_id: str, dataset_name: str, version: str
    ) -> typing.Optional[pd.DataFrame]:
        return self._load_entity(
            run_id=run_id,
            dataset_name=dataset_name,
            entity_name=TabularDatasetSerDe._ACTUALS_ENTITY_NAME,
            deserializer=_load_series,
            version=version,
        )

    def load_predictions(
        self, run_id: str, dataset_name: str, version: str
    ) -> typing.Optional[pd.DataFrame]:
        return self._load_entity(
            run_id=run_id,
            dataset_name=dataset_name,
            entity_name=TabularDatasetSerDe._PREDICTIONS_ENTITY_NAME,
            deserializer=_load_series,
            version=version,
        )



from mlflow.utils.rest_utils import MlflowHostCreds, http_request_safe

AUTH_SERVER_URL = "https://auth-server.tfy-ctl-us-east-1-develop.develop.truefoundry.io"


class AuthService:
    def __init__(self, auth_server_url: str = AUTH_SERVER_URL):
        self.host_creds = MlflowHostCreds(host=auth_server_url)

    def get_token(self, api_key: str, tenant_id: str) -> str:
        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint="/api/v1/oauth/api-keys/token",
            method="post",
            json={"apiKey": api_key, "clientId": tenant_id},
        )
        response = response.json()
        token = response["accessToken"]
        return token



import atexit
import datetime
import json
import logging
import os
import time
import typing

import mlflow
import numpy as np
import pandas as pd
import streamlit_gradio as gr
import whylogs
from mlflow.entities import Metric, Param, RunStatus, RunTag
from mlflow.tracking import MlflowClient

from mlfoundry import amplitude, constants, enums
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.frameworks import get_model_registry
from mlfoundry.git_info import GitInfo
from mlfoundry.internal_namespace import NAMESPACE
from mlfoundry.metrics.v1 import get_metrics_calculator as get_metrics_calculator_v1
from mlfoundry.metrics.v2 import ComputedMetrics
from mlfoundry.metrics.v2 import get_metrics_calculator as get_metrics_calculator_v2
from mlfoundry.run_utils import (
    NumpyEncoder,
    ParamsType,
    log_artifact_blob,
    process_params,
)
from mlfoundry.schema import Schema

log = logging.getLogger(__name__)


class MlFoundryRun:
    # TODO (nikunjbjj): Seems like these artifact locations are defined in constants.py for loading.
    #  Need to change this.
    S3_MODELS_PATH = "models"
    S3_WEBAPP_PATH = "webapp"
    S3_DATASET_PATH = "datasets"
    S3_STATS_PATH = "stats"
    S3_WHYLOGS_PATH = "whylogs"
    S3_METRICS_PATH = "multi_dimensional_metrics"

    def __init__(
        self,
        experiment_id,
        run_id,
    ):
        self.experiment_id = str(experiment_id)
        self.experiment = mlflow.get_experiment(self.experiment_id)
        self.run_id = run_id
        self.mlflow_client = MlflowClient()
        self.model_uri = "runs:/" + str(self.run_id) + "/" + self.S3_MODELS_PATH
        atexit.register(self._terminate_run_if_running, RunStatus.FINISHED)

    def __enter__(self):
        return self

    def __repr__(self) -> str:
        return f"Run ID: {self.run_id}"

    def _terminate_run_if_running(self, termination_status: RunStatus):
        # we do not need to set any termination status unless
        # the run was in RUNNING state
        try:
            current_status = self.mlflow_client.get_run(self.run_id).info.status
            if current_status != RunStatus.to_string(RunStatus.RUNNING):
                return
            self.mlflow_client.set_terminated(
                self.run_id, RunStatus.to_string(termination_status)
            )
        except Exception as ex:
            log.warning(
                f"failed to set termination status {termination_status} due to {ex}"
            )

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = RunStatus.FINISHED if exc_type is None else RunStatus.FAILED
        self._terminate_run_if_running(status)

    def end(self):
        self._terminate_run_if_running(RunStatus.FINISHED)

    def _add_git_info(self, root_path: typing.Optional[str] = None):
        root_path = root_path or os.getcwd()
        try:
            git_info = GitInfo(root_path)
            tags = [
                RunTag(
                    key=constants.GIT_COMMIT_TAG_NAME,
                    value=git_info.current_commit_sha,
                ),
                RunTag(
                    key=constants.GIT_BRANCH_TAG_NAME,
                    value=git_info.current_branch_name,
                ),
                RunTag(key=constants.GIT_DIRTY_TAG_NAME, value=str(git_info.is_dirty)),
            ]
            remote_url = git_info.remote_url
            if remote_url is not None:
                tags.append(RunTag(key=constants.GIT_REMOTE_URL_NAME, value=remote_url))
            self.mlflow_client.log_batch(run_id=self.run_id, tags=tags)
            log_artifact_blob(
                mlflow_client=self.mlflow_client,
                run_id=self.run_id,
                blob=git_info.diff_patch,
                file_name=constants.PATCH_FILE_NAME,
                artifact_path=constants.PATCH_FILE_ARTIFACT_DIR,
            )
        except Exception as ex:
            # no-blocking
            log.warning(f"failed to log git info because {ex}")

    def download_artifact(
        self, path: str, dst_path: typing.Optional[str] = None
    ) -> str:
        """Download recursively an artifact file or directory from a run to a local directory if applicable, and
           return a local path for it.
        Args:
        path : Relative source path to the desired artifact.
        dst_path : Absolute path of the local filesystem destination directory to which to download the specified artifacts.
                   This directory must already exist. If unspecified, the artifacts will either be downloaded to a new
                   uniquely-named directory on the local filesystem or will be returned directly in the case of the
                   LocalArtifactRepository.
        Returns:
        Local path of desired artifact
        """
        if dst_path is None:
            return self.mlflow_client.download_artifacts(self.run_id, path=path)
        elif os.path.isdir(dst_path):
            return self.mlflow_client.download_artifacts(
                self.run_id, path=path, dst_path=dst_path
            )
        else:
            raise MlFoundryException(
                f"Destination path {dst_path} should be an existing directory."
            )

    def __get_dataset_path(
        self, data_slice: enums.DataSlice, fileformat: enums.FileFormat = None
    ):
        if fileformat:
            return (
                str(self.run_id) + "_" + data_slice.value + "." + str(fileformat.value)
            )  # with extension
        else:
            return str(self.run_id) + "_" + data_slice.value  # without extension

    def log_artifact(self, local_path: str, artifact_path: typing.Optional[str] = None):
        if artifact_path is not None:
            NAMESPACE.validate_namespace_not_used(path=artifact_path)
        if os.path.isfile(local_path):
            self.mlflow_client.log_artifact(
                self.run_id, local_path=local_path, artifact_path=artifact_path
            )
        elif os.path.isdir(local_path):
            self.mlflow_client.log_artifacts(
                self.run_id, local_dir=local_path, artifact_path=artifact_path
            )
        else:
            raise MlFoundryException(
                f"local path {local_path} should be an existing file or directory"
            )

    def log_model(self, model, framework: typing.Union[enums.ModelFramework, str]):
        """logs a model to s3 given a model object along with the framework
        Args:
            model : model object in a particular framework
            framework (str): one of the framework
        Example:
        >> log_model(sklearn_model, ModelFramework.SKLEARN)
        """
        framework = enums.ModelFramework(framework)

        amplitude.track(
            amplitude.Event.LOG_MODEL,
            event_properties={"model_framework": framework.value},
        )
        try:
            artifacts = self.mlflow_client.list_artifacts(self.run_id)
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        for artifact in artifacts:
            if MlFoundryRun.S3_MODELS_PATH in artifact.path:
                raise MlFoundryException(f"Model already logged, cannot overwrite")

        # TODO: why do we need to do end_run and start_run?
        # this is a bug.
        mlflow.end_run()
        with mlflow.start_run(
            experiment_id=self.experiment_id,
            run_id=self.run_id,
            tags={"modelFramework": framework.value},
        ) as run:
            model_framework_class_instance = get_model_registry(framework)
            model_framework_class_instance.log_model(
                model, artifact_path=MlFoundryRun.S3_MODELS_PATH
            )

        log.info("Model logged Successfully")

    def log_dataset(
        self,
        df,
        data_slice: typing.Union[enums.DataSlice, str],
        fileformat: typing.Union[enums.FileFormat, str] = "parquet",
    ):
        """Logs a dataset to s3,
        We can only log one dataset for a slice per run.
        If attempt to log dataset more than once an error is thrown out.
        Args:
            df: Dataset to upload
            data_slice: data slice   type of ‘df’ argument
            fileformat: (Default - FileFormat.PARQUET) - format of the file for the dataset to be saved
        Example:
        >> log_dataset(dataset_df, DataSlice.TRAIN, FileFormat.PARQUET)
        """
        data_slice = enums.DataSlice(data_slice)
        fileformat = enums.FileFormat(fileformat)

        amplitude.track(amplitude.Event.LOG_DATASET)

        if not isinstance(df, pd.DataFrame):
            raise MlFoundryException("Only pandas dataframe is supported")

        artifact_name = self.__get_dataset_path(data_slice, fileformat)

        artifacts = self.mlflow_client.list_artifacts(self.run_id)
        for artifact in artifacts:
            if self.__get_dataset_path(data_slice) in artifact.path:
                raise MlFoundryException(
                    f"Dataset {str(self.run_id)}_{data_slice.value} already logged, cannot overwrite"
                )

        constants.RUN_DATASET_FOLDER.mkdir(parents=True, exist_ok=True)
        filename = os.path.join(constants.RUN_DATASET_FOLDER, artifact_name)

        if fileformat == enums.FileFormat.PARQUET:
            df.to_parquet(filename, index=False)

        elif fileformat == enums.FileFormat.CSV:
            df.to_csv(filename, index=False)

        try:
            self.mlflow_client.log_artifact(self.run_id, filename)
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        os.remove(filename)

    def log_metrics(self, metric_dict: dict, step: int = 0):
        """Logs the metrics given metric_dict that has metric name as key and metric value as value
        Args:
            metric_dict (dict): metric_dict that has metric name as key and metric value as value
            step (int): step associated with the metrics present in metric_dict
        Examples:
        >> log_metric({'accuracy':0.91, 'f1_score':0.5})
        """
        # not sure about amplitude tracking here.
        # as the user can use this function in training loop
        # amplitude.track(amplitude.Event.LOG_METRICS)

        try:
            # mlfowclient doesnt have log_metrics api, so we have to use log_batch, This is what internally used by mlflow.log_metrics
            timestamp = int(time.time() * 1000)
            metrics_arr = [
                Metric(key, value, timestamp, step=step)
                for key, value in metric_dict.items()
            ]
            if len(metrics_arr) == 0:
                raise MlflowException("Cannot log empty metrics dictionary")

            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=metrics_arr, params=[], tags=[]
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        log.info("Metrics logged successfully")

    def log_params(self, param_dict: ParamsType):
        """Logs the parameter given param_dict that has param name as key and param value as value
        Args:
            param_dict (dict): param_dict that has param name as key and param value as value
        Examples:
        >> log_params({'learning_rate':0.01, 'n_epochs:10'})

        We can also pass argparse Namspace objects directly

        ```
        import argparse
        parser = argparse.ArgumentParser()
        parser.add_argument("-batch_size", type=int, required=True)
        args = parser.parse_args()
        log_params(args)
        ```
        """
        amplitude.track(amplitude.Event.LOG_PARAMS)

        try:
            # mlfowclient doesnt have log_params api, so we have to use log_batch, This is what internally used by mlflow.log_params
            param_dict = process_params(param_dict)
            params_arr = [Param(key, str(value)) for key, value in param_dict.items()]

            if len(params_arr) == 0:
                raise MlflowException("Cannot log empty params dictionary")

            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=[], params=params_arr, tags=[]
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        log.info("Parameters logged successfully")

    def set_tags(self, tags: typing.Dict[str, typing.Any]):
        """Logs the given tags. Converts tag values to strings.
        Args:
            tags (dict): string tag name as key and tag value as value
        Examples:
        >> set_tags({'nlp.framework': 'Spark NLP'})
        """
        amplitude.track(amplitude.Event.SET_TAGS)

        try:
            NAMESPACE.validate_namespace_not_used(names=tags.keys())
            tags_arr = [RunTag(key, str(value)) for key, value in tags.items()]
            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=[], params=[], tags=tags_arr
            )
        except MlflowException as e:
            raise MlFoundryException(e.message) from e
        log.debug("Set tags successfully")

    def get_tags(self) -> typing.Dict[str, str]:
        amplitude.track(amplitude.Event.GET_TAGS)

        run = self.mlflow_client.get_run(self.run_id)
        return run.data.tags

    def __compute_whylogs_stats(self, df):

        if not isinstance(df, pd.DataFrame):
            raise MlFoundryException(
                f"df is expexted to be pd.DataFrame but got {str(type(df))}"
            )

        profile_file_name = (
            "profile"
            + "_"
            + datetime.datetime.now().strftime(constants.TIME_FORMAT)
            + ".bin"
        )
        session = whylogs.get_or_create_session()
        profile = session.new_profile()
        profile.track_dataframe(df)
        profile.write_protobuf(profile_file_name)

        try:
            self.mlflow_client.set_tag(self.run_id, "whylogs", True)
            self.mlflow_client.log_artifact(
                self.run_id,
                profile_file_name,
                artifact_path=MlFoundryRun.S3_WHYLOGS_PATH,
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        if os.path.exists(profile_file_name):
            os.remove(profile_file_name)

    def auto_log_metrics(
        self,
        model_type: enums.ModelType,
        data_slice: enums.DataSlice,
        predictions: typing.Collection[typing.Any],
        actuals: typing.Optional[typing.Collection[typing.Any]] = None,
        class_names: typing.Optional[typing.List[str]] = None,
        prediction_probabilities=None,
    ) -> ComputedMetrics:
        metrics_calculator = get_metrics_calculator_v2(model_type)
        metrics = metrics_calculator.compute_metrics(
            predictions=predictions,
            actuals=actuals,
            prediction_probabilities=prediction_probabilities,
            class_names=class_names,
        )
        metric_path = os.path.join(constants.ALM_ARTIFACT_PATH, data_slice.value)
        log_artifact_blob(
            mlflow_client=self.mlflow_client,
            run_id=self.run_id,
            blob=metrics.json(),
            file_name=constants.ALM_METRICS_FILE_NAME,
            artifact_path=metric_path,
        )
        return metrics

    # TODO (nikunjbjj): This function is too long. Need to be broken into smaller testable modules.
    def log_dataset_stats(
        self,
        df,
        data_slice: enums.DataSlice,
        data_schema: Schema,
        model_type: enums.ModelType,
        shap_values=None,
    ):
        data_slice = enums.DataSlice(data_slice)
        model_type = enums.ModelType(model_type)

        if not isinstance(df, pd.DataFrame):
            raise MlFoundryException(f"Expected pd.DataFrame but got {str(type(df))}")

        if not data_schema.actual_column_name:
            raise MlFoundryException(f"Schema.actual_column_name cannot be None")
        elif not data_schema.prediction_column_name:
            raise MlFoundryException(f"Schema.prediction_column_name cannot be None")
        elif data_schema.feature_column_names is None:
            raise MlFoundryException(f"Schema.feature_column_names cannot be None")
        elif not isinstance(data_schema.feature_column_names, list):
            raise MlFoundryException(
                f"data_schema.feature_column_names should be of type list, cannot be {type(data_schema.feature_column_names)}"
            )

        self.__compute_whylogs_stats(df[set(data_schema.feature_column_names)])

        if model_type in [
            enums.ModelType.BINARY_CLASSIFICATION,
            enums.ModelType.MULTICLASS_CLASSIFICATION,
        ]:
            class_names = None
            prediction_col_dtype, actual_col_dtype = (
                df[data_schema.prediction_column_name].dtype,
                df[data_schema.actual_column_name].dtype,
            )
            if prediction_col_dtype == object and actual_col_dtype != object:
                raise MlflowException(
                    "Both predictions column and actual column has to be of same datatype, either string or number"
                )
            elif prediction_col_dtype != object and actual_col_dtype == object:
                raise MlflowException(
                    "Both predictions column and actual column has to be of same datatype, either string or number"
                )
            elif prediction_col_dtype == object and actual_col_dtype == object:
                actual_class_names = df[data_schema.actual_column_name].unique()
                prediction_class_name = df[data_schema.prediction_column_name].unique()
                class_names = sorted(
                    set(actual_class_names) | set(prediction_class_name)
                )
                df[data_schema.actual_column_name] = df[
                    data_schema.actual_column_name
                ].apply(lambda x: class_names.index(x))
                df[data_schema.prediction_column_name] = df[
                    data_schema.prediction_column_name
                ].apply(lambda x: class_names.index(x))

        unique_count_dict = {}
        if model_type in [
            enums.ModelType.BINARY_CLASSIFICATION,
            enums.ModelType.MULTICLASS_CLASSIFICATION,
        ]:
            unique_count_dict[data_schema.prediction_column_name] = np.unique(
                df[data_schema.prediction_column_name].to_list(), return_counts=True
            )
            unique_count_dict[data_schema.actual_column_name] = np.unique(
                df[data_schema.actual_column_name].to_list(), return_counts=True
            )
        elif model_type == enums.ModelType.REGRESSION:
            session = whylogs.get_or_create_session()
            profile = session.new_profile()
            profile.track_dataframe(
                df[[data_schema.actual_column_name, data_schema.prediction_column_name]]
            )
            unique_count_dict[
                constants.ACTUAL_PREDICTION_COUNTS
            ] = profile.flat_summary()["hist"]

        if data_schema.categorical_feature_column_names:
            for feature in data_schema.categorical_feature_column_names:
                unique_count_dict[feature] = np.unique(
                    df[feature].to_list(), return_counts=True
                )

        unique_count_name = "unique_count" + "_" + str(data_slice.value) + ".json"
        constants.RUN_STATS_FOLDER.mkdir(parents=True, exist_ok=True)
        unique_count_path = os.path.join(constants.RUN_STATS_FOLDER, unique_count_name)

        with open(unique_count_path, "w") as fp:
            json.dump(unique_count_dict, fp, cls=NumpyEncoder)

        schema_json_name = "schema" + "_" + str(data_slice.value) + ".json"
        schema_json_path = os.path.join(constants.RUN_STATS_FOLDER, schema_json_name)

        with open(schema_json_path, "w") as outfile:
            json.dump(data_schema.__dict__, outfile)

        # TODO (nikunjbjj): This class name could be referenced before assignment. We need to fix this.
        if (
            model_type
            in [
                enums.ModelType.BINARY_CLASSIFICATION,
                enums.ModelType.MULTICLASS_CLASSIFICATION,
            ]
            and class_names is not None
        ):
            class_names_path = f"class_names_{data_slice.value}.json"
            class_names_path = constants.RUN_STATS_FOLDER / class_names_path
            class_names_dict = {"class_names": class_names}
            with open(class_names_path, "w") as fp:
                json.dump(class_names_dict, fp)

        metrics_class = get_metrics_calculator_v1(model_type)

        if data_schema.prediction_probability_column_name:
            metrics_dict = metrics_class.compute_metrics(
                df[set(data_schema.feature_column_names)],
                df[data_schema.prediction_column_name].to_list(),
                df[data_schema.actual_column_name].to_list(),
                df[data_schema.prediction_probability_column_name].to_list(),
            )
        else:
            metrics_dict = metrics_class.compute_metrics(
                df[set(data_schema.feature_column_names)],
                df[data_schema.prediction_column_name].to_list(),
                df[data_schema.actual_column_name].to_list(),
            )
        # non-multi dimensional metrics
        metrics_dict_with_data_slice = {}

        for key in metrics_dict[constants.NON_MULTI_DIMENSIONAL_METRICS].keys():
            new_key = "pre_computed_" + key + "_" + str(data_slice.value)
            metrics_dict_with_data_slice[new_key] = metrics_dict[
                constants.NON_MULTI_DIMENSIONAL_METRICS
            ][key]

        if shap_values is not None:
            tag_key = "data_stats_and_shap_" + data_slice.value
        else:
            tag_key = "data_stats_" + data_slice.value

        self.mlflow_client.set_tag(self.run_id, "modelType", model_type.value)
        self.mlflow_client.set_tag(self.run_id, tag_key, True)
        self.log_metrics(metrics_dict_with_data_slice)

        constants.RUN_METRICS_FOLDER.mkdir(parents=True, exist_ok=True)
        multi_dimension_metric_file = (
            "pre_computed_"
            + constants.MULTI_DIMENSIONAL_METRICS
            + "_"
            + str(data_slice.value)
            + ".json"
        )
        multi_dimension_metric_file_path = os.path.join(
            constants.RUN_METRICS_FOLDER, multi_dimension_metric_file
        )

        with open(multi_dimension_metric_file_path, "w") as fp:
            json.dump(
                metrics_dict[constants.MULTI_DIMENSIONAL_METRICS], fp, cls=NumpyEncoder
            )

        if model_type == enums.ModelType.TIMESERIES:
            actuals_predictions_filename = (
                "actuals_predictions_" + str(data_slice.value) + ".parquet"
            )
            actuals_predictions_filepath = os.path.join(
                constants.RUN_STATS_FOLDER, actuals_predictions_filename
            )
            df[
                [data_schema.prediction_column_name, data_schema.actual_column_name]
            ].to_parquet(actuals_predictions_filepath)

        try:

            # with self.mlflow_run as run:
            self.mlflow_client.log_artifact(
                self.run_id, unique_count_path, artifact_path=MlFoundryRun.S3_STATS_PATH
            )
            self.mlflow_client.log_artifact(
                self.run_id,
                multi_dimension_metric_file_path,
                artifact_path=MlFoundryRun.S3_METRICS_PATH,
            )
            self.mlflow_client.log_artifact(
                self.run_id, schema_json_path, artifact_path=MlFoundryRun.S3_STATS_PATH
            )
            if (
                model_type
                in [
                    enums.ModelType.BINARY_CLASSIFICATION,
                    enums.ModelType.MULTICLASS_CLASSIFICATION,
                ]
                and class_names is not None
            ):
                self.mlflow_client.log_artifact(
                    self.run_id,
                    class_names_path,
                    artifact_path=MlFoundryRun.S3_STATS_PATH,
                )
            # TODO (nikunjbjj): This class name path and actuals_predictions_filepath could be referenced
            #  before assignment. We need to fix this.
            if model_type == enums.ModelType.TIMESERIES:
                self.mlflow_client.log_artifact(
                    self.run_id,
                    actuals_predictions_filepath,
                    artifact_path=MlFoundryRun.S3_STATS_PATH,
                )
                os.remove(actuals_predictions_filepath)
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        os.remove(multi_dimension_metric_file_path)
        os.remove(schema_json_path)
        os.remove(unique_count_path)

        if shap_values is not None:
            self.__log_shap_values(
                df[set(data_schema.feature_column_names)], shap_values, data_slice
            )

        log.info("Dataset stats have been successfully computed and logged")

    def __log_shap_values(self, df, shap_values: list, data_slice: enums.DataSlice):

        if not isinstance(df, pd.DataFrame):
            raise MlFoundryException(f"Expected pd.DataFrame but got {str(type(df))}")

        artifact_name = str(self.run_id) + "_" + str(data_slice.value) + "_shap.json"
        constants.RUN_STATS_FOLDER.mkdir(parents=True, exist_ok=True)
        filename = os.path.join(constants.RUN_STATS_FOLDER, artifact_name)

        shap_values_dict = {"shap_values": shap_values}

        with open(filename, "w") as fp:
            json.dump(shap_values_dict, fp, cls=NumpyEncoder)

        try:
            self.mlflow_client.log_artifact(
                self.run_id, filename, artifact_path=MlFoundryRun.S3_STATS_PATH
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        os.remove(filename)

    def __get_artifact(self, run_id, artifact_name, dest_path):
        try:
            return self.mlflow_client.download_artifacts(
                run_id, artifact_name, dest_path
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

    def get_dataset(
        self, data_slice: typing.Union[enums.DataSlice, str], parse_dates=None
    ):
        data_slice = enums.DataSlice(data_slice)

        amplitude.track(amplitude.Event.GET_DATASET)
        artifacts = self.mlflow_client.list_artifacts(self.run_id)

        dataset_file_name = None
        constants.GET_RUN_DATASET_FOLDER.mkdir(parents=True, exist_ok=True)

        for artifact in artifacts:
            if self.__get_dataset_path(data_slice) in artifact.path:
                self.__get_artifact(
                    self.run_id, artifact.path, constants.GET_RUN_DATASET_FOLDER
                )
                dataset_file_name = artifact.path
                break

        if dataset_file_name:
            ext = dataset_file_name.split(".")[-1]
            dataset_file_path = os.path.join(
                constants.GET_RUN_DATASET_FOLDER, dataset_file_name
            )

            dataset_df = None
            if ext == enums.FileFormat.PARQUET.value:
                dataset_df = pd.read_parquet(dataset_file_path)
                if parse_dates != None:
                    for column_name in parse_dates:
                        dataset_df[column_name] = pd.to_datetime(
                            dataset_df[column_name]
                        )

            elif ext == enums.FileFormat.CSV.value:
                dataset_df = pd.read_csv(dataset_file_path, parse_dates=parse_dates)

            log.info(f"Dataset saved to {dataset_file_path}")
            return dataset_df
        else:
            raise MlFoundryException("Dataset not logged!")

    def get_metrics(
        self, metric_names: typing.Optional[typing.Iterable[str]] = None
    ) -> typing.Dict[str, typing.List[Metric]]:
        """
        Returns logged metrics.

        Args:
            metric_names: a list of metric names. If not passed,
                          Then returns all metrics logged for the run.
        Returns:
            A dictionary of metric_names and the logger metrics.
            The key is metric name. Value is a list of mlflow.entities.Metric object.
            If a metric has not been used before, value will be an empty list.

        ```
        In [4]: test_run.log_metrics({"loss": 0.6}, step=1)
        2022-02-21 22:32:19.550 INFO    mlfoundry.mlfoundry_run: Metrics logged successfully

        In [5]: test_run.log_metrics({"loss": 0.3}, step=2)
        2022-02-21 22:32:27.191 INFO    mlfoundry.mlfoundry_run: Metrics logged successfully

        In [6]: test_run.get_metrics()
        Out[6]:
        {'loss': [<Metric: key='loss', step=1, timestamp=1645462939544, value=0.6>,
          <Metric: key='loss', step=2, timestamp=1645462947185, value=0.3>]}
        ```
        """

        amplitude.track(amplitude.Event.GET_METRICS)
        run = self.mlflow_client.get_run(self.run_id)
        run_metrics = run.data.metrics

        metric_names = (
            set(metric_names) if metric_names is not None else run_metrics.keys()
        )

        unknown_metrics = metric_names - run_metrics.keys()
        if len(unknown_metrics) > 0:
            log.warning(f"{unknown_metrics} metrics not present in the run")
        metrics_dict = {metric_name: [] for metric_name in unknown_metrics}
        valid_metrics = metric_names - unknown_metrics
        for metric_name in valid_metrics:
            metrics_dict[metric_name] = self.mlflow_client.get_metric_history(
                self.run_id, metric_name
            )
        return metrics_dict

    def get_params(self):
        ## Ending any run that is active to prevent mlflow from throwing error
        amplitude.track(amplitude.Event.GET_PARAMS)
        run = self.mlflow_client.get_run(self.run_id)
        return run.data.params

    def get_model(self, dest_path: str = None, **kwargs):
        amplitude.track(amplitude.Event.GET_MODEL)
        artifacts = self.mlflow_client.list_artifacts(self.run_id)
        model_exists = False
        model_file_path = None
        for artifact in artifacts:
            if MlFoundryRun.S3_MODELS_PATH in artifact.path:
                model_exists = True
                model_file_path = artifact.path
                break

        if not model_exists:
            raise MlFoundryException("Model is not logged")

        run = self.mlflow_client.get_run(self.run_id)
        run_tags = run.data.tags

        framework = enums.ModelFramework(run_tags.get("modelFramework"))

        if framework is None:

            raise MlFoundryException("Invalid Framework found")

        if framework == enums.ModelFramework.TENSORFLOW:
            self.model_uri = model_file_path
        model_framework_class_instance = get_model_registry(framework)
        return model_framework_class_instance.load_model(
            self.model_uri,
            dest_path=dest_path,
            mlflow_client=self.mlflow_client,
            run_id=self.run_id,
            **kwargs,
        )

    def webapp(self, fn, inputs, outputs):
        try:
            raw_in, raw_out, _, _ = gr.Interface(
                fn=fn, inputs=inputs, outputs=outputs
            ).get_streamlit_component(key=id(self))
            return raw_in, raw_out
        except Exception as e:
            log.error(f"Webapp could not be generated. Error {e}")

    def log_webapp_file(self, file_path):
        try:
            artifacts = self.mlflow_client.list_artifacts(self.run_id)
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        for artifact in artifacts:
            if MlFoundryRun.S3_WEBAPP_PATH in artifact.path:
                raise MlFoundryException(f"Webapp already logged, cannot overwrite")

        try:
            self.mlflow_client.log_artifact(
                self.run_id, file_path, artifact_path=MlFoundryRun.S3_WEBAPP_PATH
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

    def get_webapp_file(self):
        artifacts = self.mlflow_client.list_artifacts(self.run_id)
        webapp_file_path = None
        constants.GET_RUN_WEBAPP_FOLDER.mkdir(parents=True, exist_ok=True)

        for artifact in artifacts:
            if MlFoundryRun.S3_WEBAPP_PATH in artifact.path:
                webapp_file_path = self.__get_artifact(
                    self.run_id,
                    MlFoundryRun.S3_WEBAPP_PATH,
                    constants.GET_RUN_WEBAPP_FOLDER,
                )
                break
        if webapp_file_path:
            python_files = [
                f
                for f in os.listdir(webapp_file_path)
                if os.path.isfile(os.path.join(webapp_file_path, f))
                and f.endswith(".py")
            ]
            if len(python_files) != 1:
                MlFoundryException(
                    f"Was expecting exactly one webapp file to be logged. Got {python_files}"
                )
                return
            return os.path.join(webapp_file_path, python_files[0])
        else:
            raise MlFoundryException("Webapp not logged!")



import enum
import typing

import pandas as pd
from pydantic import BaseModel
from whylogs import DatasetProfile


class FieldType(enum.Enum):
    NUMERICAL = "NUMERICAL"
    CATEGORICAL = "CATEGORICAL"
    EMPTY = "EMPTY"


class InferredFieldType(BaseModel):
    field_type: typing.Optional[FieldType]
    pandas_inferred_dtype: str


class Schema(BaseModel):
    features: typing.Dict[str, InferredFieldType]

    # only supports scalar output at this point
    # we do not support multi-label
    actuals: typing.Optional[InferredFieldType] = None
    predictions: typing.Optional[InferredFieldType] = None


class Profiles(BaseModel):
    features: DatasetProfile
    actuals: typing.Optional[DatasetProfile] = None
    predictions: typing.Optional[DatasetProfile] = None

    class Config:
        arbitrary_types_allowed = True


class DataSet(BaseModel):
    features: pd.DataFrame
    actuals: typing.Optional[pd.Series] = None
    predictions: typing.Optional[pd.Series] = None

    class Config:
        arbitrary_types_allowed = True



import mlflow

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.frameworks.base_registry import BaseRegistry


class GluonRegistry(BaseRegistry):
    def log_model(self, model, artifact_path: str, **kwargs):
        """
        Log Gluon model
        Args:
            model : the model to be registered
            artifact_path (str): artifact_path
        """
        mlflow.gluon.log_model(model, artifact_path=artifact_path)

    def load_model(self, model_file_path: str, dest_path: str = None, **kwargs):
        """
        Load Gluon model
        """
        if "ctx" not in kwargs.keys():
            raise MlFoundryException("ctx is required")
        ctx = kwargs["ctx"]
        return mlflow.gluon.load_model(model_file_path, ctx, dest_path)



import typing
from collections import defaultdict

import orjson
from sqlalchemy import create_engine, tuple_
from sqlalchemy.engine import Engine
from sqlalchemy.orm import scoped_session, sessionmaker

from mlfoundry.exceptions import MlFoundryException

from ..inference_packet import ActualPacket, BasePacket, InferencePacket
from .abstract_inference_store import AbstractInferenceStore
from .db_models import Base, InferenceLog


def orjson_serializer(obj):
    return orjson.dumps(obj, option=orjson.OPT_SERIALIZE_NUMPY).decode()


def orjson_deserializer(serialized_obj):
    return orjson.loads(serialized_obj)


def create_sqlalchemy_engine(db_uri: str) -> Engine:
    engine = create_engine(
        db_uri, json_serializer=orjson_serializer, json_deserializer=orjson_deserializer
    )
    return engine


# NOTE: for now I am keeping the above helper function in this file,
# later we may need to create a seperate db util file to keep this functions


class SqlAlchemyInferenceStore(AbstractInferenceStore):
    def __init__(self, db_uri: str):
        self.engine: Engine = create_sqlalchemy_engine(db_uri)

        # TODO: Explore albemic later
        Base.metadata.create_all(self.engine)
        self.session = scoped_session(sessionmaker(self.engine))

    def _insert_inference(
        self, session, inference_packets: typing.List[InferencePacket]
    ):
        inference_log_objects = [
            InferenceLog(
                model_name=inference.model_name,
                model_version=inference.model_version,
                inference_id=inference.inference_id,
                features=inference.features,
                predictions=inference.predictions,
                occurred_at=inference.occurred_at,
                raw_data=inference.raw_data,
                actuals=inference.actuals,
            )
            for inference in inference_packets
        ]
        session.add_all(inference_log_objects)
        session.flush()

    def _update_actuals(self, session, actual_packets: typing.List[ActualPacket]):
        indexed_packets = {
            (packet.model_name, packet.inference_id): packet.actuals
            for packet in actual_packets
        }
        packets_present_in_db = (
            session.query(InferenceLog)
            .with_entities(
                InferenceLog.id, InferenceLog.model_name, InferenceLog.inference_id
            )
            .filter(
                tuple_(InferenceLog.model_name, InferenceLog.inference_id).in_(
                    indexed_packets.keys()
                )
            )
            # TODO: We should batch this operation instead of all().
            # there can be more than one row having a particular
            # model_name and inference_id
            .all()
        )
        packets_to_update = [
            {
                "id": packet.id,
                "actuals": indexed_packets[packet.model_name, packet.inference_id],
            }
            for packet in packets_present_in_db
        ]
        session.bulk_update_mappings(InferenceLog, packets_to_update)

    def batch_log_inference(self, inference_packets: typing.List[BasePacket]):
        grouped_packets = defaultdict(list)
        for packet in inference_packets:
            grouped_packets[type(packet)].append(packet)

        # https://github.com/sqlalchemy/sqlalchemy/issues/6519
        with self.session() as session, session.begin():
            for packet_type, packets in grouped_packets.items():
                if packet_type == InferencePacket:
                    self._insert_inference(session, packets)
                elif packet_type == ActualPacket:
                    self._update_actuals(session, packets)

    def list_inference(
        self,
        model_name: str,
        model_version: typing.Optional[str] = None,
        page: int = 0,
        page_size: typing.Optional[int] = None,
    ) -> typing.List[InferencePacket]:
        if page < 0:
            raise MlFoundryException("page cannot be neg")
        if page_size is not None and page_size < 0:
            raise MlFoundryException("page_size cannot be neg")
        if page > 0 and page_size is None:
            raise MlFoundryException("pass page_size")

        filters = {"model_name": model_name}
        if model_version is not None:
            filters["model_version"] = model_version
        with self.session() as session:
            query = (
                session.query(InferenceLog)
                .filter_by(**filters)
                .order_by(InferenceLog.id)
                .offset(page * (page_size or 1))
            )
            if page_size:
                query = query.limit(page_size)
            inference_logs = query.all()

        inference_packets = [
            InferencePacket(
                model_name=inference.model_name,
                model_version=inference.model_version,
                inference_id=inference.inference_id,
                features=inference.features,
                predictions=inference.predictions,
                occurred_at=inference.occurred_at,
                raw_data=inference.raw_data,
                actuals=inference.actuals,
            )
            for inference in inference_logs
        ]
        return inference_packets



import logging
import typing

import numpy as np
from sklearn.metrics import (
    accuracy_score,
    classification_report,
    cohen_kappa_score,
    confusion_matrix,
    log_loss,
    precision_recall_curve,
    roc_curve,
)

from mlfoundry.enums import ModelType
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.metrics.v2.base_metrics import BaseMetrics
from mlfoundry.metrics.v2.custom_metric_types import (
    ClassificationReport,
    ConfusionMatrix,
    PrCurve,
    RocCurve,
)

log = logging.getLogger(__name__)


class MultiClassClassificationMetrics(BaseMetrics):
    ACCURACY = "accuracy"
    COHEN_KAPPA_SCORE = "cohen_kappa_score"
    CONFUSION_MATRIX = "confusion_matrix"
    LOG_LOSS = "log_loss"
    ROC_CURVE = "roc_curve"
    PR_CURVE = "precision_recall_curve"
    CLASSIFICATION_REPORT = "classification_report"

    @property
    def model_type(self) -> ModelType:
        return ModelType.MULTICLASS_CLASSIFICATION

    def get_metric_name_to_function_map(self) -> typing.Dict[str, typing.Callable]:
        return {
            self.ACCURACY: self.compute_accuracy,
            self.COHEN_KAPPA_SCORE: self.compute_cohen_kappa_score,
            self.CONFUSION_MATRIX: self.compute_confusion_matrix,
            self.LOG_LOSS: self.compute_log_loss,
            self.ROC_CURVE: self.compute_roc_curve,
            self.PR_CURVE: self.compute_pr_curve,
            self.CLASSIFICATION_REPORT: self.compute_classification_report,
        }

    def validate(
        self,
        predictions,
        actuals,
        prediction_probabilities,
        class_names,
    ):
        if prediction_probabilities is not None and class_names is None:
            raise MlFoundryException(
                "class names are required if prediction probability is passed"
            )
        if predictions is None:
            raise MlFoundryException("predictions cannot be None")
        if actuals is None:
            raise MlFoundryException("actuals cannot be none")
        if prediction_probabilities is not None:
            prediction_probabilities = np.asarray(prediction_probabilities, dtype=float)
            if prediction_probabilities.ndim != 2 and prediction_probabilities.shape[
                1
            ] != len(class_names):
                raise MlFoundryException(
                    "prediction_probabilities should have two dimension\n"
                    "the number of columns should match the number of class_names"
                )

    @staticmethod
    def compute_accuracy(actuals, predictions, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param predictions:
        :return:
        """
        return accuracy_score(actuals, predictions)

    @staticmethod
    def compute_cohen_kappa_score(actuals, predictions, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param predictions:
        :return:
        """
        return cohen_kappa_score(actuals, predictions)

    @staticmethod
    def compute_confusion_matrix(actuals, predictions, class_names, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param predictions:
        :return:
        """
        return ConfusionMatrix(
            matrix=confusion_matrix(actuals, predictions, labels=class_names).tolist(),
            class_names=class_names,
        )

    @staticmethod
    def compute_log_loss(actuals, prediction_probabilities, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param prediction_probabilities:
        :return:
        """
        if prediction_probabilities is None:
            log.info("prediction_probabilities not passed, skipping compute_log_loss")
            return None
        predictions = np.asarray(prediction_probabilities)
        return log_loss(actuals, predictions)

    @staticmethod
    def compute_roc_curve(actuals, prediction_probabilities, class_names, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param prediction_probabilities:
        :return:
        """
        if prediction_probabilities is None:
            log.info("prediction_probabilities not passed, skipping compute_roc_curve")
            return None
        result = {}

        prediction_probabilities = np.asarray(prediction_probabilities)

        for i, class_name in enumerate(class_names):
            fpr, tpr, thresholds = roc_curve(
                actuals, prediction_probabilities[:, i], pos_label=class_name
            )
            result[class_name] = RocCurve(
                fpr=fpr.tolist(),
                tpr=tpr.tolist(),
                thresholds=thresholds.tolist(),
            )

        return result

    @staticmethod
    def compute_pr_curve(actuals, prediction_probabilities, class_names, **kwargs):
        """
        Sklearn metric compute.
        :param actuals:
        :param prediction_probabilities:
        :return:
        """
        if prediction_probabilities is None:
            log.info("prediction_probabilities not passed, skipping compute_pr_curve")
            return None
        result = {}

        prediction_probabilities = np.asarray(prediction_probabilities)

        for i, class_name in enumerate(class_names):
            precision, recall, thresholds = precision_recall_curve(
                actuals, prediction_probabilities[:, i], pos_label=class_name
            )
            result[class_name] = PrCurve(
                precision=precision.tolist(),
                recall=recall.tolist(),
                thresholds=thresholds.tolist(),
            )

        return result

    @staticmethod
    def compute_classification_report(actuals, predictions, class_names, **kwargs):
        report = classification_report(
            y_true=actuals, y_pred=predictions, labels=class_names, output_dict=True
        )
        report = {
            class_name: ClassificationReport(**report[str(class_name)])
            for class_name in class_names
        }
        return report



import logging
import os

import click

logger = logging.getLogger(__name__)


def create_mlfoundry_cli():
    """Generates CLI by combining all subcommands into a main CLI and returns in
    Returns:
        function: main CLI functions will all added sub-commands
    """
    _cli = mlfoundry_cli()
    _cli.add_command(ui)
    return _cli


@click.group()
@click.version_option("1.0.0")
def mlfoundry_cli():
    """MlFoundry CLI"""
    click.secho("MlFoundry CLI", bold=True, fg="green")


@mlfoundry_cli.command(
    help="Generate MLFoundry Dashboard",
    short_help="Generate MLFoundry Dashboard",
)
@click.option(
    "-w",
    "--webapp_path",
    type=click.Path(),
    required=False,
    default="",
)
@click.option(
    "-p",
    "--path",
    type=click.Path(exists=True, dir_okay=True, readable=True),
    default=os.path.abspath("."),
)
def ui(path: str, webapp_path: str):
    if webapp_path:
        abs_webapp_path = os.path.abspath(webapp_path)
        basename = os.path.basename(abs_webapp_path)
        extension = basename.split(".")[-1]
        if extension != "py" or not os.path.isfile(abs_webapp_path):
            logger.error(f"Must be a valid python file. Received {webapp_path}")
            return
        os.system(f"mlfoundry_ui start-dashboard -p {path} -w {webapp_path}")
    else:
        os.system(f"mlfoundry_ui start-dashboard -p {path}")


@mlfoundry_cli.command(
    help="Generate Webapp Dashboard",
    short_help="Generate Webapp Dashboard",
)
@click.option(
    "-w",
    "--webapp_path",
    type=click.Path(exists=True, readable=True),
)
def webapp(webapp_path: str):
    os.system(f"mlfoundry_ui webapp -w {webapp_path}")



# https://github.com/whylabs/whylogs/blob/04f3c31baf0abd727f360805cfc6a2a1e5f53835/proto/src/summaries.proto

import enum
import typing

from pydantic import BaseModel, Field


class UniqueCountSummary(BaseModel):
    estimate: typing.Optional[float] = None
    upper: typing.Optional[float] = None
    lower: typing.Optional[float] = None


class FrequentStringItem(BaseModel):
    value: str
    estimate: str


class FrequentStringsSummary(BaseModel):
    items: typing.List[FrequentStringItem]


class FrequentDoubleItem(BaseModel):
    estimate: int
    value: float
    rank: int


class FrequentLongItem(BaseModel):
    estimate: int
    value: int
    rank: int


class FrequentNumbersSummary(BaseModel):
    doubles: typing.List[FrequentDoubleItem]
    longs: typing.List[FrequentLongItem]


class FrequentItem(BaseModel):
    estimate: int
    json_value: str


class FrequentItemsSummary(BaseModel):
    items: typing.List[FrequentItem]


@enum.unique
class InferredTypeEnum(enum.Enum):
    UNKNOWN = "UNKNOWN"
    NULL = "NULL"
    FRACTIONAL = "FRACTIONAL"
    INTEGRAL = "INTEGRAL"
    BOOLEAN = "BOOLEAN"
    STRING = "STRING"


class InferredType(BaseModel):
    type: InferredTypeEnum
    ratio: float


class SchemaSummary(BaseModel):
    inferred_type: InferredType
    type_counts: typing.Dict[str, int]


class HistogramSummary(BaseModel):
    start: typing.Optional[float] = None
    end: typing.Optional[float] = None
    width: typing.Optional[float] = None
    counts: typing.List[int]

    max: typing.Optional[float] = None
    min: typing.Optional[float] = None
    bins: typing.List[float]
    n: int


class QuantileSummary(BaseModel):
    quantiles: typing.List[float]
    quantile_values: typing.List[float]


class NumberSummary(BaseModel):
    count: int
    min: typing.Optional[float] = None
    max: typing.Optional[float] = None
    mean: typing.Optional[float] = None
    stddev: typing.Optional[float] = None

    histogram: HistogramSummary
    unique_count: UniqueCountSummary
    quantiles: QuantileSummary
    frequent_numbers: typing.Optional[FrequentNumbersSummary] = None

    is_discrete: typing.Optional[bool] = None


class CharPosSummary(BaseModel):
    character_list: str
    char_pos_map: typing.Dict[str, NumberSummary]


class StringsSummary(BaseModel):
    unique_count: UniqueCountSummary
    frequent: FrequentStringsSummary
    length: NumberSummary
    token_length: NumberSummary
    # char_pos_tracker: CharPosSummary


class Counters(BaseModel):
    count: int
    true_count: typing.Optional[int] = None
    null_count: typing.Optional[int] = None


class ColumnSummary(BaseModel):
    counters: Counters
    schema_summary: SchemaSummary = Field(alias="schema")
    number_summary: typing.Optional[NumberSummary] = None
    string_summary: typing.Optional[StringsSummary] = None
    frequent_items: typing.Optional[FrequentItemsSummary] = None
    unique_count: typing.Optional[UniqueCountSummary] = None

    class Config:
        allow_population_by_field_name = True


class DataFrameSummary(BaseModel):
    columns: typing.Dict[str, ColumnSummary] = Field(default_factory=dict)



import re
import typing

import pandas as pd

from mlfoundry.dataset.types import Schema
from mlfoundry.exceptions import MlFoundryException

DATASET_NAME_REGEX = re.compile(r"^[a-zA-Z0-9-]*$")


def validate_dataset_name(name: str):
    if not name or not isinstance(name, str):
        raise MlFoundryException(
            f"dataset name '{name}' is not a valid non-empty string"
        )
    if not DATASET_NAME_REGEX.match(name):
        raise MlFoundryException(
            f"dataset name '{name}' should only contain alphanumeric or dash"
        )


def validate_dataset(
    features: pd.DataFrame,
    predictions: typing.Optional[pd.Series],
    actuals: typing.Optional[pd.Series],
    schema: Schema,
):
    for feature_name, inferred_field_type in schema.features.items():
        if inferred_field_type.field_type is None:
            raise MlFoundryException(
                f"Could not identify field type of feature {feature_name} "
                f"inferred pandas type is {inferred_field_type.pandas_inferred_dtype}"
            )
    num_row_features = len(features)

    if predictions is not None:
        num_row_predictions = len(predictions)
        if num_row_features != num_row_predictions:
            raise MlFoundryException(
                f"number of rows in features {num_row_features} is not equal "
                f"to number of rows in predictions {num_row_predictions}"
            )
        if schema.predictions.field_type is None:
            raise MlFoundryException(
                "Could not identify field type of predictions "
                f"inferred pandas type is {schema.predictions.pandas_inferred_dtype}"
            )

    if actuals is not None:
        num_row_actuals = len(actuals)
        if num_row_features != num_row_actuals:
            raise MlFoundryException(
                f"number of rows in features {num_row_features} is not equal "
                f"to number of rows in actuals {num_row_actuals}"
            )
        if schema.actuals.field_type is None:
            raise MlFoundryException(
                "Could not identify field type of actuals "
                f"inferred pandas type is {schema.actuals.pandas_inferred_dtype}"
            )

    if predictions is not None and actuals is not None:
        if schema.actuals.field_type != schema.predictions.field_type:
            raise MlFoundryException(
                f"predictions field type {schema.predictions.field_type} not matching with "
                f"actuals field type {schema.actuals.field_type}"
            )



import typing

import pandas as pd
from mlflow.tracking import MlflowClient

from mlfoundry.dataset.schema import build_schema
from mlfoundry.dataset.serde import TabularDatasetSerDe
from mlfoundry.dataset.stats import build_stats
from mlfoundry.dataset.types import DataSet, Profiles, Schema
from mlfoundry.dataset.validation import validate_dataset, validate_dataset_name
from mlfoundry.enums import FileFormat
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types import TabularDatasetStats


def pre_process_features(features) -> pd.DataFrame:
    try:
        return pd.DataFrame(features).reset_index(drop=True)
    except Exception as ex:
        raise MlFoundryException("could not convert features to a DataFrame") from ex


def pre_process_actuals(actuals) -> pd.Series:
    try:
        return pd.Series(actuals).reset_index(drop=True)
    except Exception as ex:
        raise MlFoundryException("could not convert actuals to a Series") from ex


def pre_process_predictions(predictions) -> pd.Series:
    try:
        return pd.Series(predictions).reset_index(drop=True)
    except Exception as ex:
        raise MlFoundryException("could not convert predictions to a Series") from ex


class TabularDataset:
    def __init__(self, mlflow_client: MlflowClient):
        self.mlflow_client: MlflowClient = mlflow_client
        self.serde: TabularDatasetSerDe = TabularDatasetSerDe(self.mlflow_client)

    def log_dataset(
        self,
        run_id: str,
        dataset_name: str,
        features,
        predictions=None,
        actuals=None,
        only_stats: bool = False,
        file_format: FileFormat = FileFormat.PARQUET,
    ) -> str:
        validate_dataset_name(dataset_name)

        features = pre_process_features(features)
        predictions = (
            predictions if predictions is None else pre_process_predictions(predictions)
        )
        actuals = actuals if actuals is None else pre_process_actuals(actuals)

        schema = build_schema(
            features=features, predictions=predictions, actuals=actuals
        )
        validate_dataset(
            features=features, predictions=predictions, actuals=actuals, schema=schema
        )
        dataset_stats, profiles = build_stats(
            features=features, predictions=predictions, actuals=actuals
        )
        return self.serde.save_dataset(
            run_id=run_id,
            dataset_name=dataset_name,
            features=features,
            predictions=predictions,
            actuals=actuals,
            schema=schema,
            stats=dataset_stats,
            profiles=profiles,
            file_format=file_format,
            only_stats=only_stats,
        )

    def get_dataset(
        self, run_id: str, dataset_name: str, version: typing.Optional[str] = None
    ) -> DataSet:
        version = version or self.serde.get_latest_version(
            run_id=run_id, dataset_name=dataset_name
        )
        return DataSet(
            features=self.get_features(run_id, dataset_name, version=version),
            actuals=self.get_actuals(run_id, dataset_name, version=version),
            predictions=self.get_predictions(run_id, dataset_name, version=version),
        )

    def get_features(
        self, run_id: str, dataset_name: str, version: typing.Optional[str] = None
    ) -> typing.Optional[pd.DataFrame]:
        version = version or self.serde.get_latest_version(
            run_id=run_id, dataset_name=dataset_name
        )
        return self.serde.load_features(
            run_id=run_id, dataset_name=dataset_name, version=version
        )

    def get_predictions(
        self, run_id: str, dataset_name: str, version: typing.Optional[str] = None
    ) -> typing.Optional[pd.Series]:
        version = version or self.serde.get_latest_version(
            run_id=run_id, dataset_name=dataset_name
        )
        return self.serde.load_predictions(
            run_id=run_id, dataset_name=dataset_name, version=version
        )

    def get_actuals(
        self, run_id: str, dataset_name: str, version: typing.Optional[str] = None
    ) -> typing.Optional[pd.Series]:
        version = version or self.serde.get_latest_version(
            run_id=run_id, dataset_name=dataset_name
        )
        return self.serde.load_actuals(
            run_id=run_id, dataset_name=dataset_name, version=version
        )

    def get_schema(
        self, run_id: str, dataset_name: str, version: typing.Optional[str] = None
    ) -> Schema:
        version = version or self.serde.get_latest_version(
            run_id=run_id, dataset_name=dataset_name
        )
        return self.serde.load_schema(
            run_id=run_id, dataset_name=dataset_name, version=version
        )

    def get_stats(
        self, run_id: str, dataset_name: str, version: typing.Optional[str] = None
    ) -> TabularDatasetStats:
        version = version or self.serde.get_latest_version(
            run_id=run_id, dataset_name=dataset_name
        )
        return self.serde.load_stats(
            run_id=run_id, dataset_name=dataset_name, version=version
        )

    def get_profiles(
        self, run_id: str, dataset_name: str, version: typing.Optional[str] = None
    ) -> Profiles:
        version = version or self.serde.get_latest_version(
            run_id=run_id, dataset_name=dataset_name
        )
        return self.serde.load_profiles(
            run_id=run_id, dataset_name=dataset_name, version=version
        )


if __name__ == "__main__":
    import random

    import mlfoundry

    client = mlfoundry.get_client()
    # client = mlfoundry.get_client("http://localhost:5000")
    run = client.create_run("dataset-test")

    data = pd.DataFrame()
    num_rows = 10000
    data["a"] = [random.randint(-100, 100) for _ in range(num_rows)] + [None] * (
        num_rows // 4
    )
    data["b"] = [random.randint(-5, 5) for _ in range(num_rows)] + [None] * (
        num_rows // 4
    )

    data["c"] = [random.uniform(-100, 100) for _ in range(num_rows)] + [None] * (
        num_rows // 4
    )

    data["d"] = [random.choice([True, False]) for _ in range(num_rows)] + [None] * (
        num_rows // 4
    )

    data["e"] = [random.choice(["high", "low", "medium"]) for _ in range(num_rows)] + [
        None
    ] * (num_rows // 4)

    data["f"] = None
    # data.drop(data.index, inplace=True)

    dataset = TabularDataset(run.mlflow_client)
    print(
        dataset.log_dataset(
            run_id=run.run_id,
            dataset_name="my-dataset",
            features=data,
            actuals=data["a"],
            predictions=data["b"],
        )
    )
    print(
        dataset.log_dataset(
            run_id=run.run_id,
            dataset_name="my-dataset",
            features=data,
            actuals=data["a"],
            predictions=None,
        )
    )
    print(dataset.get_dataset(run_id=run.run_id, dataset_name="my-dataset"))
    print(dataset.get_schema(run_id=run.run_id, dataset_name="my-dataset"))
    print(dataset.get_stats(run_id=run.run_id, dataset_name="my-dataset"))
    print(dataset.get_profiles(run_id=run.run_id, dataset_name="my-dataset"))



class BaseRegistry:
    def log_model(self, model, artifact_path: str, **kwargs):
        return NotImplementedError

    def load_model(self, model_file_path: str, **kwargs):
        return NotImplementedError



import os
import shutil
from pathlib import Path

import mlflow

from mlfoundry.constants import GET_RUN_TMP_FOLDER, RUN_TMP_FOLDER
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.frameworks.base_registry import BaseRegistry
from mlfoundry.run_utils import download_artifact


class TensorflowRegistry(BaseRegistry):
    def log_model(self, python_object, artifact_path: str, **kwargs):
        """
        Log Tensorflow model
        Args:
            python_object : the object to be registered
            artifact_path (str): artifact_path
        """

        model_file_path = Path(os.path.join(RUN_TMP_FOLDER, artifact_path))
        model_file_path.mkdir(exist_ok=True, parents=True)

        try:
            import tensorflow as tf

            TF2 = tf.__version__.startswith("2")
        except ImportError:
            raise ImportError(
                "Tensorflow package is required to use TfSavedModelArtifact."
            )

        if TF2:
            tf.saved_model.save(
                python_object["model"],
                str(model_file_path),
                signatures=python_object["signatures"],
                options=python_object["options"],
            )
        else:
            tf.saved_model.save(
                python_object,
                str(model_file_path),
                signatures=python_object["signatures"],
            )

        mlflow.log_artifacts(model_file_path, artifact_path=artifact_path)
        shutil.rmtree(model_file_path, ignore_errors=False)

    def load_model(self, model_file_path: str, dest_path: str, **kwargs):
        """
        Load Tensorflow model
        """
        if "mlflow_client" not in kwargs.keys():
            raise MlFoundryException("mlflow_client is required")

        if "run_id" not in kwargs.keys():
            raise MlFoundryException("run_id is required")

        mlflow_client = kwargs["mlflow_client"]
        run_id = kwargs["run_id"]

        if dest_path:
            model_file_path = download_artifact(
                mlflow_client, run_id, model_file_path, dest_path
            )
        else:
            raise MlFoundryException(
                "Please provide a location 'dest_path' to save the model"
            )

        GET_RUN_TMP_FOLDER.mkdir(parents=True, exist_ok=True)
        options = kwargs.get("options", None)

        try:
            import tensorflow as tf

            TF2 = tf.__version__.startswith("2")
        except ImportError:
            raise ImportError(
                "Tensorflow package is required to use TfSavedModelArtifact."
            )

        return tf.saved_model.load(model_file_path, options=options)



import typing

import pandas as pd

from mlfoundry.dataset.types import FieldType, InferredFieldType, Schema

CATEGORICAL_PANDAS_TYPES = {"categorical", "boolean", "integer", "string"}
NUMERICAL_PANDAS_TYPES = {"integer", "floating", "mixed-integer-float"}
EMPTY_TYPE = "empty"


def infer_field_type(data: typing.Iterable) -> InferredFieldType:
    inferred_dtype = pd.api.types.infer_dtype(data)
    if inferred_dtype == EMPTY_TYPE:
        return InferredFieldType(
            field_type=FieldType.EMPTY, pandas_inferred_dtype=inferred_dtype
        )

    if inferred_dtype in NUMERICAL_PANDAS_TYPES:
        return InferredFieldType(
            field_type=FieldType.NUMERICAL, pandas_inferred_dtype=inferred_dtype
        )
    if inferred_dtype in CATEGORICAL_PANDAS_TYPES:
        return InferredFieldType(
            field_type=FieldType.CATEGORICAL, pandas_inferred_dtype=inferred_dtype
        )

    # unknown type
    return InferredFieldType(field_type=None, pandas_inferred_dtype=inferred_dtype)


def build_schema(
    features: pd.DataFrame,
    predictions: typing.Optional[pd.Series],
    actuals: typing.Optional[pd.Series],
) -> Schema:
    features_schema = dict()
    for col_name in features.columns:
        features_schema[col_name] = infer_field_type(features[col_name])

    prediction_schema = None
    if predictions is not None:
        prediction_schema = infer_field_type(predictions)

    actuals_schema = None
    if actuals is not None:
        actuals_schema = infer_field_type(actuals)

    return Schema(
        features=features_schema, actuals=actuals_schema, predictions=prediction_schema
    )



"""
# TO independently test this module, you can run the example in the path
python examples/sklearn/iris_train.py

Besides running pytest
"""
import datetime
import logging
import os
import typing
import urllib
from functools import partial

import mlflow
import pandas as pd
from mlflow.entities import ViewType
from mlflow.store.artifact.artifact_repository_registry import (
    _artifact_repository_registry,
)
from mlflow.tracking import MlflowClient
from mlflow.tracking._tracking_service.utils import _get_default_host_creds

from mlfoundry import amplitude, constants
from mlfoundry.artifact.truefoundry_artifact_repo import TruefoundryArtifactRepository
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.inference.store import (
    ActualPacket,
    InferencePacket,
    InferenceStoreClient,
    ValueType,
    get_inference_store,
)
from mlfoundry.internal_namespace import NAMESPACE
from mlfoundry.mlfoundry_run import MlFoundryRun
from mlfoundry.session import Session
from mlfoundry.tracking.auth_service import AuthService
from mlfoundry.tracking.truefoundry_rest_store import TruefoundryRestStore

log = logging.getLogger(__name__)


def init_rest_tracking(tracking_uri: str, api_key: typing.Optional[str]):
    get_cred = partial(_get_default_host_creds, tracking_uri)
    rest_store = TruefoundryRestStore(get_cred)
    auth_service = AuthService()

    session = Session(auth_service=auth_service, tracking_service=rest_store)
    session.init_session(api_key)

    artifact_repository = partial(TruefoundryArtifactRepository, rest_store=rest_store)
    _artifact_repository_registry.register("s3", artifact_repository)


def get_client(
    tracking_uri: typing.Optional[str] = None,
    inference_store_uri: typing.Optional[str] = None,
    disable_analytics: bool = False,
    api_key: typing.Optional[str] = None,
):
    amplitude.init(disable_analytics)
    if tracking_uri is None:
        tracking_uri = "file:" + os.path.abspath(constants.MLRUNS_FOLDER_NAME)
    elif tracking_uri.startswith("file:"):
        tracking_uri = os.path.join(tracking_uri, constants.MLRUNS_FOLDER_NAME)
    else:
        init_rest_tracking(tracking_uri=tracking_uri, api_key=api_key)
    amplitude.track(
        amplitude.Event.GET_CLIENT,
        # tracking whether user is using file:// or https://
        event_properties={
            "tracking_scheme": urllib.parse.urlparse(tracking_uri).scheme
        },
    )
    return MlFoundry(tracking_uri, inference_store_uri=inference_store_uri)


class MlFoundry:
    def __init__(
        self,
        tracking_uri: typing.Optional[str] = None,
        inference_store_uri: typing.Optional[str] = None,
    ):
        try:
            mlflow.set_tracking_uri(tracking_uri)
        except MlflowException as e:
            err_msg = (
                f"Could not initialise mlfoundry object. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

        self.mlflow_client = MlflowClient()
        self.inference_store_client: typing.Optional[InferenceStoreClient] = (
            InferenceStoreClient(lambda: get_inference_store(inference_store_uri))
            if inference_store_uri is not None
            else None
        )

    def _get_or_create_experiment(
        self, experiment_name: str, owner: typing.Optional[str]
    ):
        """
        Creates a experiment and returns the experiment id. This experiment is to be passed into run.
        :param
            experiment_name: A unique experiment name
            owner (Optional: str): owner of the project. If owner is not passed,
                                   the current user will be used as owner. If the given owner
                                   does not have the project, it will be created under
                                   the current user.

        :return: The experiment_id of created experiment

        Example:
        >> create_experiment('experiment-1')
        """

        try:
            experiment = self.mlflow_client.get_experiment_by_name(
                experiment_name, owner_subject_id=owner
            )
            if experiment is not None:
                return experiment.experiment_id
            if not owner:
                log.info(
                    f"project {experiment_name} does not exist. Creating {experiment_name}."
                )
                return self.mlflow_client.create_experiment(experiment_name)
            else:
                log.warning(
                    f"project {experiment_name} under owner {owner} does not exist. "
                    "looking for project under current user."
                )
                return self._get_or_create_experiment(
                    experiment_name=experiment_name, owner=None
                )
        except MlflowException as e:
            err_msg = (
                f"Error happened in creating or getting experiment based on experiment name: "
                f"{experiment_name}. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

    def get_all_projects(self):
        """
        Returns names of all the projects
        """
        amplitude.track(amplitude.Event.GET_ALL_PROJECTS)
        try:
            experiments = self.mlflow_client.list_experiments(view_type=ViewType.ALL)
        except MlflowException as e:
            err_msg = (
                f"Error happened in fetching project names. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

        projects = []
        for e in experiments:
            # Experiment ID 0 represents default project which we are removing.
            if e.experiment_id != "0":
                projects.append(e.name)

        return projects

    def rename_project(self, old_project_name: str, new_project_name: str):
        """
        Renames a project.
        :param old_project_name: Existing project name
        :param new_project_name: New Project name
        """

        amplitude.track(amplitude.Event.RENAME_PROJECT)
        try:
            experiment_id = self.mlflow_client.get_experiment_by_name(
                old_project_name
            ).experiment_id

            self.mlflow_client.rename_experiment(experiment_id, new_project_name)
        except MlflowException as e:
            err_msg = (
                f"Error happened in renaming project from {old_project_name} to "
                f"{new_project_name}. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

    def create_run(
        self,
        project_name: str,
        run_name: str = constants.DEFAULT_RUN_NAME,
        tags: typing.Optional[typing.Dict[str, typing.Any]] = None,
        owner: typing.Optional[str] = None,
    ) -> MlFoundryRun:
        """Creates a run for the given project_name(project_name).
        Each run will have a unique run_id.
        If run_name is not provided, then a name is generated automatically.
        Args:
            project_name (str): name of a project
            run_name (Optional[str]): name of the run
            tags (Optional: Dict[str, Any]): Tags to associate with this run
            owner (Optional: str): owner of the project. If owner is not passed,
                                   the current user will be used as owner. If the given owner
                                   does not have the project, it will be created under
                                   the current user.
        Returns:
            MlFoundryRun: an MlFoundryRun with specified experiment and unique run_id

        Example:
        >> mlf_run = create_run('my_project')
        """

        amplitude.track(amplitude.Event.CREATE_RUN)
        if not run_name:
            run_name = constants.DEFAULT_RUN_NAME
        if project_name == "" or (not isinstance(project_name, str)):
            raise MlFoundryException(
                f"project_name must be string type and not empty. "
                f"Got {type(project_name)} type with value {project_name}"
            )

        experiment_id = self._get_or_create_experiment(project_name, owner=owner)

        if tags is not None:
            NAMESPACE.validate_namespace_not_used(tags.keys())
        else:
            tags = {}

        run = self.mlflow_client.create_run(experiment_id, name=run_name, tags=tags)
        mlf_run_id = run.info.run_id

        mlf_run = MlFoundryRun(experiment_id, mlf_run_id)
        mlf_run._add_git_info()
        log.info(f"Run is created with id {mlf_run_id} and name {run_name}")
        return mlf_run

    def get_run(self, run_id: str) -> MlFoundryRun:
        """Given the run_id returns the Python MlFoundryRun object that is created already.

        Args:
            run_id (str): run_id that was created already.

        Returns:
            MlFoundryRun: returns the run object that was created already.

        Example:
        >> run = get_run(<run_id>)
        """
        amplitude.track(amplitude.Event.GET_RUN)
        if run_id == "" or (not isinstance(run_id, str)):
            raise MlFoundryException(
                f"run_id must be string type and not empty. "
                f"Got {type(run_id)} type with value {run_id}"
            )

        run = self.mlflow_client.get_run(run_id)
        experiment_id = run.info.experiment_id
        return MlFoundryRun(experiment_id, run.info.run_id)

    def get_all_runs(self, project_name: str, owner: typing.Optional[str] = None):
        """Returns all the run that was created by the user under the project project_name
        Args:
            project_name (str): name of the project
            owner (Optional: str): owner of the project. If owner is not passed,
                                   the current user will be used as owner.

        Returns:
            pd.DataFrame: dataframe with two columns- run_id and run_name
        """
        amplitude.track(amplitude.Event.GET_ALL_RUNS)
        if project_name == "" or (not isinstance(project_name, str)):
            raise MlFoundryException(
                f"project_name must be string type and not empty. "
                f"Got {type(project_name)} type with value {project_name}"
            )
        experiment = self.mlflow_client.get_experiment_by_name(
            project_name, owner_subject_id=owner
        )
        if experiment is None:
            return pd.DataFrame(
                columns=[constants.RUN_ID_COL_NAME, constants.RUN_NAME_COL_NAME]
            )

        experiment_id = experiment.experiment_id

        try:
            all_run_infos = self.mlflow_client.list_run_infos(
                experiment_id, run_view_type=ViewType.ALL
            )
        except MlflowException as e:
            err_msg = f"Error happened in while fetching runs for project {project_name}. Error details: {e.message}"
            raise MlFoundryException(err_msg) from e

        runs = []

        for run_info in all_run_infos:
            try:
                run = self.mlflow_client.get_run(run_info.run_id)
                run_name = run.info.name or run.data.tags.get(
                    constants.RUN_NAME_COL_NAME, ""
                )
                runs.append((run_info.run_id, run_name))
            except MlflowException as e:
                log.warning(
                    f"Could not fetch details of run with run_id {run_info.run_id}. "
                    f"Skipping this one. Error details: {e.message}. "
                )

        return pd.DataFrame(
            runs, columns=[constants.RUN_ID_COL_NAME, constants.RUN_NAME_COL_NAME]
        )

    @staticmethod
    def get_tracking_uri():
        return mlflow.tracking.get_tracking_uri()

    def log_prediction(
        self,
        model_name: str,
        model_version: str,
        inference_id: str,
        features: ValueType,
        predictions: ValueType,
        raw_data: typing.Optional[ValueType] = None,
        actuals: typing.Optional[ValueType] = None,
        occurred_at: typing.Optional[int] = None,
    ):
        if self.inference_store_client is None:
            raise MlFoundryException(
                "Pass inference_store_uri in get_client function to use log_prediction"
            )
        if occurred_at is None:
            occurred_at = datetime.datetime.utcnow()
        elif not isinstance(occurred_at, int):
            raise TypeError("occurred_at should be unix epoch")
        else:
            occurred_at = datetime.datetime.utcfromtimestamp(occurred_at)
        inference_packet = InferencePacket(
            model_name=model_name,
            model_version=model_version,
            features=features,
            predictions=predictions,
            inference_id=inference_id,
            raw_data=raw_data,
            actuals=actuals,
            occurred_at=occurred_at,
        )
        self.inference_store_client.log_predictions([inference_packet])

    def log_actuals(self, model_name: str, inference_id: str, actuals: ValueType):
        if self.inference_store_client is None:
            raise MlFoundryException(
                "Pass inference_store_uri in get_client function to use log_prediction"
            )
        actuals_packet = ActualPacket(
            model_name=model_name, inference_id=inference_id, actuals=actuals
        )
        self.inference_store_client.log_actuals([actuals_packet])



import logging

from mlfoundry.constants import (
    MULTI_DIMENSIONAL_METRICS,
    NON_MULTI_DIMENSIONAL_METRICS,
    PROB_MULTI_DIMENSIONAL_METRICS,
    PROB_NON_MULTI_DIMENSIONAL_METRICS,
)

log = logging.getLogger(__name__)


class BaseMetrics:
    PERCENTAGE_FEATURE_NULL = "percentage_feature_null"

    def __init__(self):
        self.metrics_base = {
            NON_MULTI_DIMENSIONAL_METRICS: {},
            MULTI_DIMENSIONAL_METRICS: {},
            PROB_NON_MULTI_DIMENSIONAL_METRICS: {},
            PROB_MULTI_DIMENSIONAL_METRICS: {},
        }

    def compute_metrics(
        self, features_df, predictions, labels, pred_probabilities=None
    ):
        """
        Compute metrics for all models.
        :param features_df:
        :param predictions:
        :param labels:
        :param pred_probabilities:
        :return:
            metrics_dict
        """

        all_metrics = {
            NON_MULTI_DIMENSIONAL_METRICS: {},
            MULTI_DIMENSIONAL_METRICS: {},
            PROB_NON_MULTI_DIMENSIONAL_METRICS: {},
            PROB_MULTI_DIMENSIONAL_METRICS: {},
        }
        for metric in self.metrics_base[NON_MULTI_DIMENSIONAL_METRICS].keys():
            all_metrics[NON_MULTI_DIMENSIONAL_METRICS][metric] = self.metrics_base[
                NON_MULTI_DIMENSIONAL_METRICS
            ][metric]()
        for metric in self.metrics_base[MULTI_DIMENSIONAL_METRICS].keys():
            all_metrics[MULTI_DIMENSIONAL_METRICS][metric] = self.metrics_base[
                MULTI_DIMENSIONAL_METRICS
            ][metric]()

        if pred_probabilities:
            for metric in self.metrics_base[PROB_NON_MULTI_DIMENSIONAL_METRICS].keys():
                all_metrics[NON_MULTI_DIMENSIONAL_METRICS][metric] = self.metrics_base[
                    PROB_NON_MULTI_DIMENSIONAL_METRICS
                ][metric]()
            for metric in self.metrics_base[PROB_MULTI_DIMENSIONAL_METRICS].keys():
                all_metrics[MULTI_DIMENSIONAL_METRICS][metric] = self.metrics_base[
                    PROB_MULTI_DIMENSIONAL_METRICS
                ][metric]()

        return all_metrics



import logging
import os
import typing

from mlfoundry.tracking.auth_service import AuthService
from mlfoundry.tracking.truefoundry_rest_store import TruefoundryRestStore

TRACKING_TOKEN_ENV_VAR = "MLFLOW_TRACKING_TOKEN"
API_KEY_ENV_VAR = "MLF_API_KEY"

logger = logging.getLogger(__name__)


class Session:
    def __init__(
        self, auth_service: AuthService, tracking_service: TruefoundryRestStore
    ):
        self.auth_service: AuthService = auth_service
        self.tracking_service: TruefoundryRestStore = tracking_service
        self._api_key: typing.Optional[str] = None

    def _reset_session(self):
        self._api_key = None
        os.environ.pop(TRACKING_TOKEN_ENV_VAR, None)

    def init_session(
        self,
        api_key: typing.Optional[str] = None,
    ):
        self._reset_session()

        tenant_id = self.tracking_service.get_tenant_id()

        if api_key is not None:
            self._api_key = api_key
        elif API_KEY_ENV_VAR in os.environ:
            self._api_key = os.environ[API_KEY_ENV_VAR]

        if self._api_key is None:
            logger.info(
                "Session was not set as api key was neither passed, not set via env var"
            )
            return

        token = self.auth_service.get_token(api_key=self._api_key, tenant_id=tenant_id)
        os.environ[TRACKING_TOKEN_ENV_VAR] = token



import argparse
import json
import os
import tempfile
import time
import typing
from collections.abc import Mapping

import mlflow
import numpy as np

from mlfoundry.exceptions import MlflowException, MlFoundryException


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def download_artifact(mlflow_client, run_id, artifact_name, dest_path):
    try:
        return mlflow_client.download_artifacts(run_id, artifact_name, dest_path)
    except MlflowException as e:
        raise MlFoundryException(e.message).with_traceback(e.__traceback__) from None


ParamsType = typing.Union[typing.Mapping[str, typing.Any], argparse.Namespace]


def process_params(params: ParamsType) -> typing.Dict[str, typing.Any]:
    if isinstance(params, Mapping):
        return params
    if isinstance(params, argparse.Namespace):
        return vars(params)
    # TODO: add absl support if required
    # move to a different file then
    raise MlFoundryException(
        "params should be either argparse.Namespace or a Mapping (dict) type"
    )


def log_artifact_blob(
    mlflow_client: mlflow.tracking.MlflowClient,
    run_id: str,
    blob: typing.Union[str, bytes],
    file_name: str,
    artifact_path: typing.Optional[str] = None,
):
    with tempfile.TemporaryDirectory(prefix=run_id) as tmpdirname:
        local_path = os.path.join(tmpdirname, file_name)
        mode = "wb" if isinstance(blob, bytes) else "w"
        with open(local_path, mode) as local_file:
            local_file.write(blob)
        mlflow_client.log_artifact(run_id, local_path, artifact_path=artifact_path)


def mapping_to_mlflow_metric(
    metrics: typing.Mapping[str, float],
    timestamp: typing.Optional[int] = None,
    step: int = 0,
) -> typing.Dict[str, mlflow.entities.Metric]:
    if timestamp is None:
        timestamp = int(time.time() * 1000)
    mlflow_metrics = {
        key: mlflow.entities.Metric(key, value, timestamp, step=step)
        for key, value in metrics.items()
    }
    return mlflow_metrics
